from enum import Enum, Flag
from typing import Any

__all__ = (
    'ResponseType',
    'RequestType',
    'ButtonStyle',
    'CommandType',
    'OptionType',
    'ChannelType',
    'ComponentType'
)

class ResponseType(Enum):
    """
    Outgoing response types
    """
    PONG = 1 # shouldn't really ever be used

    CHANNEL_MESSAGE_WITH_SOURCE = 4 # sending an initial message
    SEND_MESSAGE = 4

    DEFERRED_CHANNEL_MESSAGE_WITH_SOURCE = 5 # deferring a slash command
    DEFER = 5

    DEFERRED_UPDATE_MESSAGE = 6 # deferring a message component
    COMPONENT_DEFER = 6

    UPDATE_MESSAGE = 7 # editing the original slash command message
    EDIT_ORIGINAL_MESSAGE = 7

    APPLICATION_COMMAND_AUTOCOMPLETE_RESULT = 8 # sending an autocomplete result
    AUTOCOMPLETE = 8

    MODAL = 9 # sending a modal form

class ButtonStyle(Enum):
    """
    Enum for the different button styles.
    """
    PRIMARY = 1
    BLURPLE = 1

    SECONDARY = 2
    GREY = 2
    GRAY = 2

    SECCESS = 3
    GREEN = 3

    DANGER = 4
    RED = 4

    LINK = 5
    URL = 5

class CommandType(Enum):
    """
    Enum for the different types of commands.
    """
    CHAT_INPUT = 1 # Slash Command
    USER = 2
    MESSAGE = 3

class RequestType(Enum):
    """
    Incoming discord interaction types
    """
    PING = 1
    APPLICATION_COMMAND = 2
    MESSAGE_COMPONENT = 3
    APPLICATION_COMMAND_AUTOCOMPLETE = 4
    MODAL_SUBMIT = 5

class OptionType(Enum):
    """
    Enum for the different types of options.
    """
    SUB_COMMAND = 1
    SUB_COMMAND_GROUP = 2
    STRING = 3
    INTEGER = 4
    BOOLEAN = 5
    USER = 6
    CHANNEL = 7
    ROLE = 8
    MENTIONABLE = 9
    NUMBER = 10

class ChannelType(Enum):
    """
    Enum for the different types of channels.
    """
    GUILD_TEXT = 0
    DM = 1
    GUILD_VOICE = 2
    GROUP_DM = 3
    GUILD_CATEGORY = 4
    GUILD_NEWS = 5
    GUILD_STORE = 6
    GUILD_NEWS_THREAD = 10
    GUILD_PUBLIC_THREAD = 11
    GUILD_PRIVATE_THREAD = 12
    GUILD_STAGE_VOICE = 13

class ComponentType(Enum):
    """
    Enum for the different types of components.
    """
    ACTION_ROW = 1
    BUTTON = 2
    SELECT = 3
    INPUT_TEXT = 4

class TextStyleTypes(Enum):
    SHORT = 1
    PARAGRAPH = 2

class Permissions:
    """
    Enum for the different permissions.
    """
    CREATE_INSTANT_INVITE = 1
    KICK_MEMBERS = 1 << 1
    BAN_MEMBERS = 1 << 2
    ADMINISTRATOR = 1 << 3
    MANAGE_CHANNELS = 1 << 4
    MANAGE_GUILD = 1 << 5
    ADD_REACTIONS = 1 << 6
    VIEW_AUDIT_LOG = 1 << 7
    PRIORITY_SPEAKER = 1 << 8
    STREAM = 1 << 9
    VIEW_CHANNEL = 1 << 10
    SEND_MESSAGES = 1 << 11
    SEND_TTS_MESSAGES = 1 << 12
    MANAGE_MESSAGES = 1 << 13
    EMBED_LINKS = 1 << 14
    ATTACH_FILES = 1 << 15
    READ_MESSAGE_HISTORY = 1 << 16
    MENTION_EVERYONE = 1 << 17
    USE_EXTERNAL_EMOJIS = 1 << 18
    VIEW_GUILD_INSIGHTS = 1 << 19
    CONNECT = 1 << 20
    SPEAK = 1 << 21
    MUTE_MEMBERS = 1 << 22
    DEAFEN_MEMBERS = 1 << 23
    MOVE_MEMBERS = 1 << 24
    USE_VAD = 1 << 25
    CHANGE_NICKNAME = 1 << 26
    MANAGE_NICKNAMES = 1 << 27
    MANAGE_ROLES = 1 << 28
    MANAGE_WEBHOOKS = 1 << 29
    MANAGE_EMOJIS_AND_STICKERS = 1 << 30
    USE_APPLICATION_COMMANDS = 1 << 31
    REQUEST_TO_SPEAK = 1 << 32
    MANAGE_EVENTS = 1 << 33
    MANAGE_THREADS = 1 << 34
    CREATE_PUBLIC_THREADS = 1 << 35
    CREATE_PRIVATE_THREADS = 1 << 36
    USE_EXTERNAL_STICKERS = 1 << 37
    SEND_MESSAGES_IN_THREADS = 1 << 38
    USE_EMBEDDED_ACTIVITIES = 1 << 39
    MODERATE_MEMBERS = 1 << 40

    def __init__(self, value):
        self.value = int(value)

    def __or__(self, other):
        return Permissions(self.value | other.value)

    def __and__(self, other):
        return Permissions(self.value & other.value)

    def __xor__(self, other):
        return Permissions(self.value ^ other.value)
    
    def __invert__(self):
        return Permissions(~self.value)

    def __repr__(self):
        return f'Permissions({self.value})'

    def __str__(self):
        return f'Permissions({self.value})'

    def __int__(self):
        return self.value

    def __bool__(self):
        return self.value != 0

    def __eq__(self, other):
        return self.value == other.value

    def __ne__(self, other):
        return self.value != other.value

    def __hash__(self):
        return hash(self.value)

    def __contains__(self, other):
        return self.value & other.value == other.value

    def __iter__(self):
        for perm in Permissions:
            if self.value & perm.value:
                yield perm

    def __len__(self):
        return len(list(self))

    def __instancecheck__(self, __instance: Any) -> bool:
        return isinstance(__instance, Permissions)



__all__ = (
    'Color',
    'Colour',
)

class Color:
    """
    Color class for embeds.
    """

    __slots__ = ('value',)

    r = property(lambda self: self.value >> 16 & 0xFF)
    g = property(lambda self: self.value >> 8 & 0xFF)
    b = property(lambda self: self.value & 0xFF)

    def __init__(self, value: int):
        if not isinstance(value, int):
            raise TypeError(f'Expected int parameter, received {value.__class__.__name__} instead.')

        self.value: int = value

    def __str__(self) -> str:
        return f'#{self.value:0>6x}'

    def __int__(self) -> int:
        return self.value

    def __repr__(self) -> str:
        return f'<Color value={self.value}>'

    @classmethod
    def from_rgb(cls, r: int, g: int, b: int):
        """
        Creates a color from RGB values.
        """
        return cls(int(r) << 16 | int(g) << 8 | int(b))

    @classmethod
    def from_hex(cls, hex_code: str):
        """
        Creates a color from a hex code.
        """
        return cls.from_rgb(*[int(hex_code[i:i+2], 16) for i in (0, 2, 4)])

Colour = Color


from dataclasses import asdict
from typing import Callable, Coroutine, List, Optional, Union, Iterable
from abc import ABC, abstractmethod
import asyncio

from .embed import Embed
from .models import Choice
from .enums import ResponseType
from .components import Components, Button, Select, TextInput

MISSING = object()

__all__ = (
    'AutocompleteResponse',
    'MessageResponse',
    'DeferredResponse',
    'EditResponse',
    'ModalResponse',
)

class _DiscordResponse(ABC):
    def __init__(
        self,
        type: ResponseType,
        **kwargs
    ) -> None:
        self.type = type
        self.data = kwargs

    @abstractmethod
    def to_dict(self):
        return {
            "type": self.type.value,
            "data": self.data
        }

class AutocompleteResponse(_DiscordResponse):
    def __init__(
        self,
        *choices: List[Choice],
        **kwargs
    ) -> None:
        super().__init__(ResponseType.AUTOCOMPLETE, choices=choices, **kwargs)

    def to_dict(self):
        return {
            "type": self.type.value,
            "data": {
                "choices": [asdict(x) for x in self.data["choices"]]
            }
        }

class DeferredResponse(_DiscordResponse):
    def __init__(
        self,
        task: Union[asyncio.Task, Callable] = None,
        ephemeral: bool = False,
        **kwargs
    ) -> None:

        kwargs['flags'] = 64 * int(ephemeral)

        self.task = task

        super().__init__(ResponseType.DEFER, **kwargs)

    def to_dict(self):
        return super().to_dict()

class MessageResponse(_DiscordResponse):
    def __init__(
        self, 
        content: str = MISSING,
        embed: Embed = MISSING,
        embeds: List[Embed] = MISSING,
        components: Components | list[Button | Select] = MISSING,
        ephemeral: bool = False,
        type: ResponseType = ResponseType.SEND_MESSAGE,
    ) -> None:
        self.type = type
        self.content = content
        self.embed = embed
        self.embeds = embeds
        self.ephemeral = ephemeral
        self.components = MISSING

        if components is not MISSING:
            if isinstance(components, list):
                for c in components:
                    self.add_component(c)
            elif isinstance(components, Components):
                self.components = components
            elif isinstance(components, (Button, Select)):
                self.add_component(components)
            else:
                raise TypeError(f"components must be Components or a list of Button and Select, not {components.__class__}")

    def add_component(self, component: Union[Button, Select], row: int = None):
        if isinstance(component, TextInput):
            raise ValueError("TextInput cannot be added to a message")

        if self.components is MISSING:
            self.components = Components()
        self.components.add_component(component, row)

        return self

    def remove_component(self, component: Union[Button, Select, str]):
        if self.components is MISSING:
            self.components = Components()
        self.components.remove_component(component)

        return self

    def add_embed(self, embed: Embed):
        if self.embeds is MISSING:
            self.embeds = []

        if len(self.embeds) >= 10:
            raise ValueError("Cannot add more than 10 embeds to a message")
        self.embeds.append(embed)

        return self

    def to_dict(self):
        if self.type is ResponseType.PONG:
            return {
                "type": self.type.value,
            }

        data = {}

        if self.embeds is not MISSING:
            data['embeds'] = [e.to_dict() for e in self.embeds]

        if self.embed is not MISSING:
            if 'embeds' not in data:
                data['embeds'] = []
            data['embeds'].append(self.embed.to_dict())

        if self.components is not MISSING:
            data['components'] = self.components.to_dict()

        if self.content is not MISSING:
            data['content'] = self.content

        if self.ephemeral:
            data['flags'] = 64

        return {
            "type": self.type.value,
            "data": data
        }

class EditResponse(MessageResponse):
    def __init__(self, *args, **kwargs) -> None:
        kwargs['type'] = ResponseType.EDIT_ORIGINAL_MESSAGE
        if 'ephemeral' in kwargs:
            del kwargs['ephemeral']
        super().__init__(**kwargs)

    def to_dict(self):
        return super().to_dict()

class ModalResponse(_DiscordResponse):
    def __init__(
        self,
        custom_id: str,
        title: str,
        components: Components | list[TextInput] = MISSING,
    ) -> None:
        self.custom_id = custom_id
        self.title = title
        self.components = MISSING
        
        if components is not MISSING:
            if isinstance(components, list):
                for c in components:
                    self.add_component(c)
            elif isinstance(components, Components):
                self.components = components
            elif isinstance(components, TextInput):
                self.add_component(components)
            else:
                raise TypeError(f"components must be Components or a list of TextInput, not {components.__class__}")

    def add_component(self, component: TextInput, row: int = None):
        if not isinstance(component, TextInput):
            raise ValueError("Modals only support TextInput components")

        if self.components is MISSING:
            self.components = Components()
        self.components.add_component(component, row)

        return self

    def remove_component(self, component: Union[TextInput, str], row: int = None):
        if self.components is MISSING:
            self.components = Components()
        self.components.remove_component(component)

        return self

    def to_dict(self):
        return {
            "type": ResponseType.MODAL.value,
            "data": {
                "custom_id": self.custom_id,
                "title": self.title,
                "components": self.components.to_dict()
            }
        }


import functools
import inspect
from pydoc import cli
from typing import Optional

from snowfin.decorators import Interactable, InteractionCommand, ComponentCallback, ModalCallback, Listener

__all__ = (
    'Module',
)

class Module:

    auto_defer = None,
    description: Optional[str] = None,
    enabled: bool = True

    def __new__(
        cls,
        client,
        *args, **kwargs
    ):
        new_cls = super().__new__(cls)

        new_cls.app = client.app
        new_cls.client = client
        new_cls.http = client.http

        new_cls.callbacks = []
        new_cls.description = cls.description or cls.__doc__

        for _name, val in inspect.getmembers(
            new_cls, predicate=lambda x: isinstance(x, Interactable)
        ):
            val.module = new_cls

            if not isinstance(val.callback, functools.partial):
                val_name = val.__name__
                val.callback = functools.partial(val.callback, new_cls)
                val.callback.__name__ = val_name

            new_cls.callbacks.append(val)

            if isinstance(val, InteractionCommand):
                client.add_interaction_command(val)
            elif isinstance(val, Listener):
                client.add_listener(val)
            elif isinstance(val, ComponentCallback):
                client.add_component_callback(val)
            elif isinstance(val, ModalCallback):
                client.add_modal_callback(val)

        client.log(f"Loaded {cls.__name__} with {len(new_cls.callbacks)} callbacks")

        return new_cls

    def on_unload(self):
        pass

    def on_load(self):
        pass


        

    



"""
snowfin
~~~~~~~~~~~~~~~~~~~
An async discord http interactions framework built on top of Sanic
:copyright: (c) 2021-present kaj
:license: MIT, see LICENSE for more details.
"""

__title__ = 'snowfin'
__author__ = 'kajdev'
__license__ = 'MIT'
__copyright__ = 'Copyright 2021-present kaj'
__version__ = '1.0'

from .decorators import *
from .response import *
from .embed import *
from .client import *
from .errors import *
from .color import *
from .enums import *
from .components import *
from .module import *
from .models import *

# hi mom


import asyncio
from dataclasses import dataclass, field, asdict
from typing import Callable, Optional, Union

from snowfin.enums import ChannelType, CommandType, ComponentType, OptionType
from .models import Choice

__all__ = (
    'SlashCommand',
    'ComponentCallback',
    'ModalCallback',
    'SlashOption',
    'ContextMenu',
    'Listener',
    'slash_command',
    'slash_option',
    'context_menu',
    'message_command',
    'user_command',
    'listen',
    'component_callback',
    'select_callback',
    'button_callback',
    'modal_callback',
)

@dataclass
class Interactable:
    callback: Optional[Callable] = None
    module = None

    def __call__(self, *args, **kwargs):
        return self.callback(*args, **kwargs)

    def __str__(self):
        return f'{self.module}:{self.callback.__name__}'

    @property
    def __name__(self) -> str:
        return self.callback.__name__

@dataclass
class FollowupMixin:
    after_callback: Optional[Callable] = None

    def followup(self) -> Callable:
        def wrapper(callback):
            self.after_callback = callback
            return callback

        return wrapper

@dataclass 
class CustomIdMappingsMixin:
    """
    A mixin for converting mappigs within interaction custom ids

    e.g. the custom_id "role:{id}:{user}" provided with the mappings:
    {'id': int, 'user': int} will add the `id` and `user` kwargs with
    converted values if possible.

    example usage:

    ```
    # this will match a custom_id like "add_role:123" and pass the value int(123)
    @button_callback("add_role:{role}")
    async def add_role_via_button(ctx, role: int):
        pass
    ```
    """
    mappings: dict = field(default_factory=dict)
    chopped_id: list[str] = field(default_factory=list)

@dataclass
class InteractionCommand(Interactable, FollowupMixin):
    """
    Discord command
    """
    name: str = None
    default_permission: bool = True

@dataclass
class ComponentCallback(Interactable, FollowupMixin, CustomIdMappingsMixin):
    """
    Discord component callback
    """
    custom_id: str = None
    type: ComponentType = None

@dataclass
class ModalCallback(Interactable, FollowupMixin, CustomIdMappingsMixin):
    """
    Discord modal callback
    """
    custom_id: str = None

@dataclass
class Listener(Interactable):
    """
    Discord listener
    """
    event_name: str = None

@dataclass
class SlashOption:
    """
    Discord command option
    """
    name: str
    type: OptionType
    description: str
    min_value: Optional[Union[int, float]] = None
    max_value: Optional[Union[int, float]] = None
    choices: Optional[list[Choice]] = None
    options: Optional[list['SlashOption']] = None
    channel_types: Optional[list[ChannelType]] = None
    required: bool = False
    autocomplete: bool = False

    def to_dict(self):
        d = {
            'name': self.name,
            'type': self.type.value if isinstance(self.type, OptionType) else self.type,
            'description': self.description,
            'min_value': self.min_value,
            'max_value': self.max_value,
            'required': self.required,
            'autocomplete': self.autocomplete
        }

        if self.choices:
            d['choices'] = [c.to_dict() if isinstance(c, Choice) else c for c in self.choices]

        if self.options:
            d['options'] = [o.to_dict() if isinstance(o, SlashOption) else o for o in self.options]

        if self.channel_types:
            d['channel_types'] = [c.value if isinstance(c, ChannelType) else c for c in self.channel_types]

        return d

@dataclass
class SlashCommand(InteractionCommand):
    name: str = None
    description: str = "No Description Set"

    options: list[SlashOption | dict] = field(default_factory=list)
    autocomplete_callbacks: dict = field(default_factory=dict)

    def __post_init__(self):
        if self.options:
            self.options = [SlashOption(**o) if isinstance(o, dict) else o for o in self.options]

        if self.callback is not None:
            if hasattr(self.callback, 'options'):
                if not self.options:
                    self.options = []
                self.options += self.callback.options

    def to_dict(self):
        d = {
            'name': self.name,
            'description': self.description,
            'type': CommandType.CHAT_INPUT.value,
            'default_permission': self.default_permission
        }

        if self.options:
            d['options'] = [o.to_dict() if isinstance(o, SlashOption) else o for o in self.options]

        return d

    def autocomplete(self, option_name: str) -> Callable:
        def wrapper(callback):
            self.autocomplete_callbacks[option_name] = callback

            # set the autocomplete value in the corresponding option
            if self.options is None:
                self.options = []
            for option in self.options:
                if option.name == option_name:
                    option.autocomplete = True
                    break
            else:
                raise ValueError(f"Option {option_name} not found")

            return callback

        return wrapper

@dataclass
class ContextMenu(InteractionCommand):
    type: CommandType = None

    def to_dict(self):
        return {
            'name': self.name,
            'type': self.type.value,
            'default_permission': self.default_permission
        }

def slash_command(
    name: str,
    description: str = None,
    options: list[SlashOption] = None,
    default_permission: bool = True,
    **kwargs
) -> Callable:
    """
    Create a slash command
    """
    def wrapper(callback):
        if not asyncio.iscoroutinefunction(callback):
            raise ValueError("Commands must be coroutines")

        return SlashCommand(
            name=name,
            description=description or callback.__doc__ or "No Description Set",
            options=options or [],
            default_permission=default_permission,
            callback=callback,
            **kwargs
        )
    
    return wrapper

def slash_option(
    name: str,
    description: str,
    type: OptionType,
    min_value: Optional[int | float] = None,
    max_value: Optional[int | float] = None,
    choices: Optional[list[Choice]] = None,
    options: Optional[list[SlashOption]] = None,
    channel_types: Optional[list[ChannelType]] = None,
    required: bool = False,
    autocomplete: bool = False,
) -> Callable:
    """
    Create a slash option
    """
    def wrapper(callback):
        if not asyncio.iscoroutinefunction(callback):
            raise ValueError("Commands must be coroutines")

        option = SlashOption(
            name=name,
            description=description,
            type=type,
            min_value=min_value,
            max_value=max_value,
            choices=choices if choices else [],
            options=options,
            channel_types=channel_types,
            required=required,
            autocomplete=autocomplete,
        )

        if not hasattr(callback, 'options'):
            callback.options = []
        callback.options.insert(0, option)

        return callback
    
    return wrapper

def context_menu(
    name: str,
    type: CommandType,
    default_permission: bool = True,
    **kwargs
) -> Callable:
    """
    Create a context menu
    """
    def wrapper(callback):
        if not asyncio.iscoroutinefunction(callback):
            raise ValueError("Commands must be coroutines")

        return ContextMenu(
            name=name,
            type=type,
            default_permission=default_permission,
            callback=callback,
            **kwargs
        )
    
    return wrapper

def message_command(
    name: str,
    default_permission: bool = True,
    **kwargs
) -> Callable:
    """
    Create a message command
    """
    def wrapper(callback):
        if not asyncio.iscoroutinefunction(callback):
            raise ValueError("Commands must be coroutines")

        return ContextMenu(
            name=name,
            type=CommandType.MESSAGE,
            default_permission=default_permission,
            callback=callback,
            **kwargs
        )
    
    return wrapper

def user_command(
    name: str,
    default_permission: bool = True,
    **kwargs
) -> Callable:
    """
    Create a user command
    """
    def wrapper(callback):
        if not asyncio.iscoroutinefunction(callback):
            raise ValueError("Commands must be coroutines")

        return ContextMenu(
            name=name,
            type=CommandType.USER,
            default_permission=default_permission,
            callback=callback,
            **kwargs
        )
    
    return wrapper

def listen(event_name: str = None) -> Callable:
    """
    Create a listener
    """
    def wrapper(callback):
        if not asyncio.iscoroutinefunction(callback):
            raise ValueError("Listeners must be coroutines")

        return Listener(
            event_name=event_name or callback.__name__,
            callback=callback
        )
    
    return wrapper

def component_callback(
    custom_id: str,
    type: ComponentType,
    __no_mappings__: bool = False,
    **kwargs
) -> Callable:
    """
    Create a component callback
    """
    def wrapper(callback):
        if not asyncio.iscoroutinefunction(callback):
            raise ValueError("Callbacks must be coroutines")

        if __no_mappings__:
            mappings = chopped_id = None
        else:
            mappings = kwargs

            chopped_id = []
            left = [custom_id]

            for kw, tp in callback.__annotations__.items():
                if (param := '{'+kw+'}') in custom_id:
                    mappings[kw] = tp
                    _, *left = ''.join(left).split(param)

                    if not _:
                        raise ValueError(f"Mapped custom_id must have characters separating the mapped parameters")

                    chopped_id.append(_)

        return ComponentCallback(
            custom_id=custom_id,
            callback=callback,
            type=type,
            mappings=mappings,
            chopped_id=chopped_id
        )
    
    return wrapper

def select_callback(
    custom_id: str,
    **kwargs
) -> Callable:
    """
    Create a select callback
    """
    return component_callback(custom_id, ComponentType.SELECT, **kwargs)

def button_callback(
    custom_id: str,
    **kwargs
) -> Callable:
    """
    Create a button callback
    """
    return component_callback(custom_id, ComponentType.BUTTON, **kwargs)

def modal_callback(
    custom_id: str,
    __no_mappings__: bool = False,
    **kwargs
) -> Callable:
    """
    Create a modal callback
    """
    def wrapper(callback):
        if not asyncio.iscoroutinefunction(callback):
            raise ValueError("Callbacks must be coroutines")

        if __no_mappings__:
            mappings = chopped_id = None
        else:
            mappings = kwargs

            chopped_id = []
            left = [custom_id]

            for kw, tp in callback.__annotations__.items():
                if (param := '{'+kw+'}') in custom_id:
                    mappings[kw] = tp
                    _, *left = ''.join(left).split(param)

                    if not _:
                        raise ValueError(f"Mapped custom_id must have characters separating the mapped parameters")

                    chopped_id.append(_)

        return ModalCallback(
            custom_id=custom_id,
            callback=callback,
            mappings=mappings,
            chopped_id=chopped_id
        )
    
    return wrapper


import asyncio
from contextlib import suppress
from dataclasses import dataclass
import functools
import importlib
import inspect
import sys
from typing import Callable, Optional
from functools import partial

from sanic import Sanic, Request
import sanic
from sanic.response import HTTPResponse, json
from sanic.log import logger

from nacl.signing import VerifyKey
from nacl.exceptions import BadSignatureError

from dacite import from_dict, config
from snowfin.components import Components, TextInput, is_component
from snowfin.errors import CogLoadError

from snowfin.models import *
from snowfin.module import Module
from .decorators import Interactable, InteractionCommand
from .response import _DiscordResponse, AutocompleteResponse, DeferredResponse, MessageResponse, ModalResponse
from .http import *
from .decorators import *
from .enums import *

from json import dumps

__all__ = (
    'Client',
    'AutoDefer',
)

cast_config=config.Config(
    cast=[
        int,
        ChannelType,
        CommandType,
        OptionType,
        ComponentType,
        RequestType
    ]
)

@dataclass
class AutoDefer:
    enabled: bool = False
    timeout: int = 1
    ephemeral: bool = False

class Client:

    def __init__(
        self,
        verify_key: str,
        application_id: int,
        sync_commands: bool = False,
        token: str = None,
        auto_defer: AutoDefer | bool = AutoDefer(),
        app: Sanic = None,
        logging_level: int = 1,
        **kwargs
    ):
        # create a new app if none is not supplied
        if app is None:
            self.app = Sanic("snowfin-interactions")
        else:
            self.app = app
        self.verify_key = VerifyKey(bytes.fromhex(verify_key))

        # automatic defer options
        self.auto_defer = auto_defer or AutoDefer()

        if self.auto_defer is True:
            self.auto_defer = AutoDefer(enabled=True)

        self.sync_commands = sync_commands

        self.log = lambda *msgs: logger.info(' '.join(str(msg) for msg in msgs))
        self.error = lambda *msgs: logger.error(' '.join(str(msg) for msg in msgs))

        self.http: HTTP = HTTP(
            application_id=application_id,
            token=token,
            proxy=kwargs.get('proxy', None),
            proxy_auth=kwargs.get('proxy_auth', None),
            headers=kwargs.get('headers', None),
        )

        self.user: User = None

        self.modules = {}

        # listeners for events (read only events)
        self._listeners: dict[str, list] = {}

        # strict callbacks (returns a response)
        self.commands: list[InteractionCommand] = []
        self.modals: dict[str, ModalCallback] = {}
        self.components: dict[tuple[str, ComponentType], ComponentCallback] = {}

        # gather callbacks
        self._gather_callbacks()

        # create some middleware for start and stop events
        @self.app.listener('after_server_start')
        async def on_start(app, loop):
            await self._sync_commands()
            self.dispatch('start')

            if self.http.application_id:
                self.user = await self.fetch_user(self.http.application_id)

        @self.app.listener('before_server_stop')
        async def on_stop(app, loop):
            self.dispatch('stop')

        # create middlware for verifying that discord is the one who sent the interaction
        @self.app.on_request
        async def verify_signature(request: Request):
            signature = request.headers.get("X-Signature-Ed25519")
            timestamp = request.headers.get("X-Signature-Timestamp")
            body = request.body.decode("utf-8")

            try:
                self.verify_key.verify(f'{timestamp}{body}'.encode(), bytes.fromhex(signature))
            except BadSignatureError:
                return json({"error": "invalid signature"}, status=403)

        # send PONGs to PINGs and construct the interaction context
        @self.app.on_request
        async def ack_request(request: Request):
            if request.json.get('type') == RequestType.PING.value:
                return json({"type": 1})

            # if self.app.debug:
            #     self.log(f"{request.method} {request.path}\n\n{dumps(request.json, indent=2)}")

            request.ctx = from_dict(
                data= request.json,
                data_class=Interaction,
                config=cast_config
            )

            request.ctx.client = self

        # handle user callbacks
        @self.app.post("/")
        async def handle_request(request: Request):
            return await self._handle_request(request)

        logger.info("Client initialized")

    @property
    def loop(self):
        try:
            return self.app.loop
        except sanic.exceptions.SanicException:
            return None
    
    async def _sync_commands(self):
        if self.sync_commands:
            type_classes = {
                CommandType.CHAT_INPUT.value: SlashCommand,
                CommandType.MESSAGE.value: ContextMenu,
                CommandType.USER.value: ContextMenu
            }

            current_commands = [
                from_dict(data=cmd, data_class=type_classes.get(cmd.get('type')), config=cast_config)
                for cmd in await self.http.get_global_application_commands()
            ]

            current_commands = [x.to_dict() for x in current_commands]
            gathered_commands = [x.to_dict() for x in self.commands]

            for cmd in gathered_commands:
                if cmd not in current_commands:
                    
                    self.log(f"syncing {len(self.commands)} commands")
                    await self.http.bulk_overwrite_global_application_commands(gathered_commands)
                    self.log(f"synced {len(self.commands)} commands")

                    return
                

    def _handle_deferred_routine(self, routine: asyncio.Task, request, after: Optional[Callable]):
        """
        Create a wrapper for the task supplied and wait on it.
        log any errors and pass the result onward
        """
        async def wrapper():
            try:
                response = await routine
                await self._handle_deferred_response(request, response)

                # call the after callback if it exists
                if after:
                    await self._handle_followup_response(request, after)
            except Exception as e:
                logger.error(e.__repr__())
        asyncio.create_task(wrapper())

    async def _handle_deferred_response(self, request, response):
        """
        Take the result of a deferred callback task and send a request to the interaction webhook
        """
        if response:
            if response.type in (ResponseType.SEND_MESSAGE, ResponseType.EDIT_ORIGINAL_MESSAGE):
                await self.http.edit_original_message(request, response)
            else:
                raise Exception("Invalid response type")

    async def _handle_followup_response(self, request, after):
        """
        Take the result of a followup and send a request to the interaction webhook
        """
        response = await after()

        if response:
            if not isinstance(response, (_DiscordResponse, HTTPResponse)):
                response = self.infer_response(response)

            if response.type is ResponseType.SEND_MESSAGE:
                await self.http.send_followup(request, response)
            elif response.type is ResponseType.EDIT_ORIGINAL_MESSAGE:
                await self.http.edit_original_message(request, response)
            else:
                raise Exception("Invalid response type")

    def infer_response(self, resp: Any) -> _DiscordResponse:
        kwargs = {}
        chosen_type = None

        if not isinstance(resp, tuple):
            resp = [resp]
            
        for arg in resp:
            if isinstance(arg, ResponseType):
                kwargs['type'] = arg

            elif isinstance(arg, Embed):
                kwargs.setdefault('embeds', []).append(arg)
                chosen_type = chosen_type or MessageResponse

            elif is_component(arg):
                kwargs.setdefault('components', Components()).add_component(arg)
                chosen_type = chosen_type or (ModalResponse if isinstance(arg, TextInput) else MessageResponse)

            elif isinstance(arg, Components):
                kwargs['components'] = arg

                for row in arg.rows:
                    for component in row.components:
                        if isinstance(component, TextInput):
                            chosen_type = chosen_type or ModalResponse
                            break
                else:
                    chosen_type = chosen_type or MessageResponse

            elif isinstance(arg, str):
                kwargs['content'] = arg
                chosen_type = chosen_type or MessageResponse

            elif isinstance(arg, list):
                kwargs.setdefault('choices', []).extend(arg)
                chosen_type = chosen_type or AutocompleteResponse

            elif isinstance(arg, dict):
                kwargs.update(arg)
            else:
                raise ValueError(f"Invalid response type {arg}")

        return (chosen_type or MessageResponse)(**kwargs)

    async def _handle_request(self, request: Request) -> HTTPResponse:
        """
        Grab the callback Coroutine and create a task.
        """
        # handle the before requests
        self.dispatch('before_request', request.ctx)

        func: Optional[Callable] = None
        after: Optional[Callable] = None

        if request.ctx.type is RequestType.APPLICATION_COMMAND:
            self.dispatch('command', request.ctx)

            if cmd := self.get_command(request.ctx.data.name):
                kwargs = {}

                for option in request.ctx.data.options:

                    converted = option.value
                    if option.type in (OptionType.CHANNEL, OptionType.USER, OptionType.ROLE, OptionType.MENTIONABLE):
                        converted = request.ctx.data.resolved.get(option.type, option.value)
                    else:
                        converted = type(option.value)

                    kwargs[option.name] = converted
                        

                func = partial(cmd.callback, request.ctx, **kwargs)
                if cmd.after_callback:
                    after = partial(cmd.after_callback, request.ctx, **kwargs)

        elif request.ctx.type is RequestType.APPLICATION_COMMAND_AUTOCOMPLETE:
            self.dispatch('autocomplete', request.ctx)

            if cmd := self.get_command(request.ctx.data.name):
                for option in request.ctx.data.options:
                    if option.focused:
                        callback = cmd.autocomplete_callbacks.get(option.name)
                        if callback:
                            func = partial(callback, request.ctx, option.value)
                        break

        elif request.ctx.type is RequestType.MESSAGE_COMPONENT:
            self.dispatch('component', request.ctx)

            func, after = self.package_component_callback(
                request.ctx.data.custom_id,
                request.ctx.data.component_type,
                request.ctx
            )

        elif request.ctx.type is RequestType.MODAL_SUBMIT:
            self.dispatch('modal', request.ctx)

            if modal := self.modals.get(request.ctx.data.custom_id):
                func = partial(modal, request.ctx)

                if modal.after_callback:
                    after = partial(modal.after_callback, request.ctx)

        if self.app.debug: self.log(f"getting callback for {request.ctx.type}: found", f"{func.func.__name__}{func.args[1:]}" if func else None)

        if func:
            task = asyncio.create_task(func())

            # auto defer if and only if the decorator and/or client told us too and it *can* be defered
            if self.auto_defer.enabled and \
                request.ctx.type in (RequestType.APPLICATION_COMMAND, RequestType.MESSAGE_COMPONENT):
                # we want to defer automatically and keep the original task going
                # so we wait for up to the timeout, then construct a DeferredResponse ourselves
                # then handle_deferred_routine() will do the rest
                done, pending = await asyncio.wait([task], timeout = self.auto_defer.timeout)


                if task in pending:
                    # task didn't return in time, let it keep going and construct a defer for it
                    resp = DeferredResponse(task,
                        ephemeral=self.auto_defer.ephemeral
                    )
                else:
                    # the task returned in time, get the result and use that like normal
                    resp = task.result()
            else:
                resp = await task
        else:
            return json({"error": "command not found"}, status=404)

        if request.ctx.responded:
            raise Exception("Callback already responded")

        if not isinstance(resp, DeferredResponse):
            request.ctx.responded = True

        if not isinstance(resp, (_DiscordResponse, HTTPResponse)):
            resp = self.infer_response(resp)

        if isinstance(resp, _DiscordResponse):
            if isinstance(resp, DeferredResponse):
                # make sure we are sending the correct interaction response type for the request
                if request.ctx.type == RequestType.MESSAGE_COMPONENT:
                    resp.type = ResponseType.COMPONENT_DEFER
                else:
                    resp.type = ResponseType.DEFER

                # if someone passed in a callable, construct a task for them to keep syntax as clean as possible
                if not isinstance(resp.task, asyncio.Task):
                    resp.task = asyncio.create_task(resp.task(self, request.ctx))

                # start or continue the task and post the response to a webhook
                self._handle_deferred_routine(resp.task, request, after)
            else:
                # launch after callbacks if there is any and the command is not a deferred one
                if after:
                    asyncio.create_task(self._handle_followup_response(request, after))
            
            # do some logging and return the 'dictified' data
            data = resp.to_dict()
            if self.app.debug: self.log(f"RESPONDING {request.ctx.type} `{getattr(request.ctx.data, 'name', None)}`", data)
            return json(data)

        elif isinstance(resp, HTTPResponse):
            # someone gave us a sanic response, Assume they know what they are doing
            return resp


    def run(self, host: str, port: int, **kwargs):
        self.app.run(host=host, port=port, access_log=False, **kwargs)

    def load_module(self, module_name: str):
        resolved_name = importlib.util.resolve_name(module_name, __spec__.parent)

        if resolved_name in self.modules:
            raise CogLoadError(f"{module_name} already loaded")

        try:
            module = importlib.import_module(module_name)
        except Exception as e:
            raise CogLoadError(f"{module_name} failed to load: {e}")
        self.modules[resolved_name] = []
        
        # go through every class that inherits from Module
        for cls in inspect.getmembers(module, inspect.isclass):
            if issubclass(cls[1], Module) and cls[1].enabled and cls[1] != Module:
                new_module = cls[1](self)
                self.modules[resolved_name].append(new_module)

                if hasattr(new_module, 'on_load'):
                    new_module.on_load()
                
    def unload_module(self, module: str):
        module_name = importlib.util.resolve_name(module, __spec__.parent)
        
        if module_name not in self.modules:
            raise ValueError(f"{module_name} not loaded")

        modules: list[Module] = self.modules.pop(module_name)

        for module in modules:
            if hasattr(module, "on_unload"):
                module.on_unload()

            # unload all callbacks
            for callback in module.callbacks:
                self.remove_callback(callback)

            del module

    def get_module(self, name: str) -> Optional[Module]:
        """
        Get a loaded module by name
        """

        for modules in self.modules.values():
            for module in modules:
                if module.__class__.__name__ == name:
                    return module

    def _ingest_callbacks(self, *callbacks: Interactable):
        for func in callbacks:
            if isinstance(func, InteractionCommand):
                self.add_interaction_command(func)
            elif isinstance(func, Listener):
                self.add_listener(func)
            elif isinstance(func, ComponentCallback):
                self.add_component_callback(func)
            elif isinstance(func, ModalCallback):
                self.add_modal_callback(func)

    def _gather_callbacks(self) -> None:
        """
        Gather all callbacks from loaded modules
        """
        callbacks = [obj for _, obj in inspect.getmembers(sys.modules["__main__"]) + inspect.getmembers(self) if isinstance(obj, Interactable)]
        self._ingest_callbacks(*callbacks)
        self.log(f"Gathered {len(callbacks)} immediate callbacks")

    def add_interaction_command(self, command: InteractionCommand):
        """
        Add a command to the client
        """
        if command.name in [x.name for x in self.commands]:
            raise ValueError(f"/{command.name} already exists")

        self.commands.append(command)

    def add_listener(self, listener: Listener):
        """
        Add a listener to the client
        """
        listener.event_name = listener.event_name.removeprefix("on_")

        if listener in self._listeners.get(listener.event_name, []):
            raise ValueError(f"{listener} already exists")

        self._listeners.setdefault(listener.event_name, []).append(listener)

    def add_component_callback(self, callback: ComponentCallback):
        """
        Add a component callback to the client
        """
        if (callback.custom_id, callback.type) in self.components:
            raise ValueError(f"{callback.type} with custom_id `{callback.custom_id}` already exists")

        self.components[(callback.custom_id, callback.type)] = callback

    def add_modal_callback(self, callback: ModalCallback):
        """
        Add a modal callback to the client
        """
        if callback.custom_id in self.modals:
            raise ValueError(f"modal with custom_id `{callback.custom_id}` already exists")

        self.modals[callback.custom_id] = callback

    def dispatch(self, event: str, *args, **kwargs) -> None:
        """
        Dispatch an event to all listeners
        """
        self.log(f"Dispatching {event}")
        for listener in self._listeners.get(event, []):
            asyncio.create_task(
                listener(*args, **kwargs),
                name=f"snowfin:: {event}"
            )

    def get_command(self, name: str) -> InteractionCommand:
        """
        Get a command by name
        """
        for command in self.commands:
            if command.name == name:
                return command

    def package_component_callback(self, custom_id: str, component_type: ComponentType, ctx: Interaction) -> Callable:
         # loop through all all our registered component callbacks
        for (_id, _type), callback in self.components.items():

            # check the type first and foremost
            if _type == component_type:

                kwargs = {}

                # make sure there are actually mappings to check
                if None not in (callback.mappings, callback.chopped_id):
                    just_values = []

                    left = custom_id

                    # go through all the constants in the defined custom_id and
                    # check if they match the mappings. Construct a list of the
                    # values to pass to the callback and convert
                    for i in range(len(callback.chopped_id)):
                        
                        # this is the next constant in the custom_id
                        segment = callback.chopped_id[i]

                        # make sure the constant is in the custom_id
                        if segment not in left:
                            break

                        # strip the constant from the custom_id so we know that
                        # the next part of the string is the value
                        left = left.removeprefix(segment)
                        if i+1 < len(callback.mappings):
                            value = left.strip(callback.chopped_id[i+1])[0]
                        else:
                            value = left
                        
                        just_values.append(value)

                        # remove the value from the custom_id so we know
                        # that the next part of the string is the next constant
                        left = left.removeprefix(value)
                            
                    # check to make sure that we have the right number of values collected
                    if len(just_values) != len(callback.mappings):
                        continue

                    mappings = callback.mappings.items()
                    for i, (name, _type) in enumerate(mappings):
                        # convert the value to the correct type if possible

                        kwargs[name] = just_values[i]

                        with suppress(ValueError):
                            kwargs[name] = _type(kwargs[name])
                elif _id != custom_id:
                        continue


                return (
                    functools.partial(callback.callback, ctx, **kwargs),
                    functools.partial(callback.after_callback, ctx, **kwargs) if callback.after_callback else None
                )

        return None, None


    def remove_callback(self, callback: Interactable):
        """
        Remove a callback from the client
        """
        if isinstance(callback, InteractionCommand):
            self.commands.remove(callback)
        elif isinstance(callback, Listener):
            self._listeners.get(callback.event_name, []).remove(callback)
        elif isinstance(callback, ComponentCallback):
            self.components.pop((callback.custom_id, callback.type))
        elif isinstance(callback, ModalCallback):
            self.modals.pop(callback.custom_id)



    async def fetch_user(self, user_id: int) -> User:
        """
        Fetch a user object
        """
        data = await self.http.fetch_user(user_id)
        if data is not None:
            return from_dict(User, data, config=cast_config)


from typing import Union, List
from .enums import ButtonStyle, ComponentType, TextStyleTypes

__all__ = (
    'Components',
    'Emoji',
    'Button',
    'Select',
    'SelectOption',
    'ActionRow',
    'TextInput',
    'is_component',
)


class Emoji:
    def __init__(self, name: str, id: int, animated: bool = False):
        self.name = name
        self.id = id
        self.animated = animated

    @classmethod
    def from_str(cls, emoji_string: str):
        if len(emoji_string) == 1:
            return cls(name=emoji_string, id=None)

        emoji_string = emoji_string.strip('<>')
        data = emoji_string.split(':')
        if len(data) == 3:
            return cls(data[1], str(data[2]), animated='a' in data[0])
        elif len(data) == 2:
            return cls(data[0], str(data[1]))
        else:
            raise ValueError(f"Invalid emoji string: {emoji_string}")

    def __str__(self):
        return f"<{'a' if self.animated else ''}:{self.name}:{self.id}>"

    def __repr__(self):
        return f"<Emoji {self.name=} {self.id=} {self.animated=}>"

    def to_dict(self):
        d = {
            'name': self.name,
            'id': self.id
        }

        if self.animated:
            d['animated'] = True

        return d

class Button:
    def __init__(
        self,
        label: str,
        custom_id: str = None,
        disabled: bool = False,
        style: ButtonStyle = ButtonStyle.PRIMARY,
        emoji: Union[str, Emoji] = None,
        url: str = None,
        **kwargs
    ) -> None:
        self.weight = 1
        self.type = ComponentType.BUTTON
        self.label = label
        self.custom_id = custom_id
        self.disabled = disabled
        self.style = ButtonStyle(style)

        if isinstance(emoji, str):
            self.emoji = Emoji.from_str(emoji)
        else:
            self.emoji = emoji

        self.url = url

        if self.style is ButtonStyle.URL and self.url is None:
            raise ValueError("URL button requires a URL")
        
        if self.url is not None:
            self.style = ButtonStyle.URL
            self.custom_id = None

    def to_dict(self):
        d = {
            "type": self.type.value,
            "label": self.label,
            "disabled": self.disabled,
            "style": self.style.value,
        }

        if self.style is ButtonStyle.URL:
            d['url'] = self.url
        else:
            d['custom_id'] = self.custom_id

        if self.style is not ButtonStyle.URL and self.custom_id is None:
            raise ValueError("Button requires a custom_id")

        if self.emoji is not None:
            d['emoji'] = self.emoji.to_dict()

        return d

class SelectOption:
    def __init__(self, label: str, value: str, description: str = None, emoji: str = None, default: bool = False) -> None:
        self.label = label
        self.value = value
        self.description = description
        self.default = default

        if isinstance(emoji, str):
            self.emoji = Emoji.from_str(emoji)
        elif isinstance(emoji, dict):
            self.emoji = Emoji(**emoji)
        else:
            self.emoji = emoji

    def to_dict(self):
        d = {
            "label": self.label,
            "value": self.value,
            "default": self.default,
        }

        if self.emoji is not None:
            d["emoji"] = self.emoji.to_dict()

        if self.description is not None:
            d["description"] = self.description

        return d

class Select:
    def __init__(
        self,
        custom_id: str,
        placeholder: str = None,
        disabled: bool = False,
        options: List[SelectOption] = None,
        min_values: int = 1,
        max_values: int = 1,
        **kwargs
    ) -> None:
        self.weight = 5
        self.type = ComponentType.SELECT
        self.custom_id = custom_id
        self.placeholder = placeholder
        self.disabled = disabled

        if len(options) == 0:
            raise ValueError("Select requires at least one option")
        elif len(options) > 25:
            raise ValueError("Select cannot have more than 25 options")
        self.options = [SelectOption(**x) if not isinstance(x, SelectOption) else x for x in options]

        self.min_values = min_values
        self.max_values = max_values

        if self.min_values > self.max_values:
            raise ValueError("min_values cannot be greater than max_values")

        if self.min_values < 0:
            self.min_values = 0
        if self.max_values > 25:
            self.max_values = 25

    def add_option(self, option: SelectOption) -> None:
        if len(self.options) >= 25:
            raise ValueError("Select cannot have more than 25 options")
        self.options.append(option)

        return self

    def remove_option(self, option: Union[int, SelectOption]) -> None:
        if isinstance(option, int):
            self.options.pop(option)
        elif isinstance(option, SelectOption):
            self.options.remove(option)
        else:
            raise ValueError("option must be an int or SelectOption")

        return self

    def to_dict(self):
        return {
            "type": self.type.value,
            "custom_id": self.custom_id,
            "placeholder": self.placeholder,
            "disabled": self.disabled,
            "options": [x.to_dict() for x in self.options],
            "min_values": self.min_values,
            "max_values": self.max_values
        }

class TextInput:
    def __init__(
        self,
        custom_id: str,
        label: str = None,
        style: TextStyleTypes = TextStyleTypes.SHORT,
        placeholder: str = None,
        min_length: int = None,
        max_length: int = None,
        **kwargs
    ) -> None:
        self.weight = 5
        self.type = ComponentType.INPUT_TEXT
        self.custom_id = custom_id
        self.label = label
        self.style = TextStyleTypes(style)
        self.placeholder = placeholder
        self.min_length = min_length
        self.max_length = max_length

    def to_dict(self):
        d = {
            "type": self.type.value,
            "custom_id": self.custom_id,
            "style": self.style.value,
            "label": self.label,
        }

        if self.placeholder is not None:
            d["placeholder"] = self.placeholder

        if self.min_length is not None:
            d["min_length"] = self.min_length

        if self.max_length is not None:
            d["max_length"] = self.max_length

        return d


class ActionRow:
    def __init__(self, *components) -> None:
        self.components = []
        self.weights = 0
        self.type = ComponentType.ACTION_ROW

        for component in components:
            self.add_component(component)

    def add_component(self, component: Union[Button, Select]):
        if self.weights + component.weight > 5:
            raise ValueError("Cannot add component, weight limit exceeded")
        self.components.append(component)
        self.weights += component.weight

        return self

    def remove_component(self, index: int):
        self.weights -= self.components[index].weight
        del self.components[index]

        return self

    def to_dict(self):
        return {
            "type": self.type.value,
            "components": [x.to_dict() for x in self.components]
        }

class Components:

    def __init__(self, *components) -> None:
        self.rows = [ActionRow() for _ in range(5)]

        for component in components:
            self.add_component(component)

    def add_component(self, component: Union[Button, Select, TextInput], row: int = None):
        if row is None:
            for _row in self.rows:
                if _row.weights + component.weight <= 5:
                    _row.add_component(component)
                    return self

            raise ValueError("Cannot add component, weight limit exceeded")
        else:
            if row > len(self.rows):
                raise ValueError("Row does not exist")
            if self.rows[row].weights + component.weight <= 5:
                self.rows[row].add_component(component)
                return self
            raise ValueError("Cannot add component, weight limit exceeded")

    def add_component_raw(self, component: dict, row: int = None):
        if component.get('type') == ComponentType.BUTTON.value:
            component = Button(**component)
        elif component.get('type') == ComponentType.SELECT.value:
            component = Select(**component)
        elif component.get('type') == ComponentType.INPUT_TEXT.value:
            component = TextInput(**component)
        else:
            raise ValueError("Component type not supported")

        return self.add_component(component, row)

    def replace_component(self, component: Union[Button, Select, str], new_component: Union[Button, Select]):
        if isinstance(component, str):
            for row in self.rows:
                for index, _component in enumerate(row.components):
                    if _component.custom_id == component and _component.type == new_component.type:
                        row[index] = new_component
                        return self
            raise ValueError("Component does not exist")
        elif isinstance(component, (Button, Select)):
            for row in self.rows:
                for index, _component in enumerate(row.components):
                    if _component == component:
                        row[index] = new_component
                        return self
            raise ValueError("Component does not exist")
        else:
            raise ValueError("component must be a custom_id, Button or Select")

    def remove_component(self, component: Union[Button, Select, str]):
        if isinstance(component, str):
            for row in self.rows:
                for index, _component in enumerate(row.components):
                    if _component.custom_id == component:
                        row.remove_component(index)
                        return self
        elif isinstance(component, (Button, Select)):
            for row in self.rows:
                for index, _component in enumerate(row.components):
                    if _component == component:
                        row.remove_component(index)
                        return self
        else:
            raise ValueError("component must be a custom_id, Button or Select")

    def to_dict(self):
        data = []
        for row in self.rows:
            if row.weights > 0:
                data.append(row.to_dict())
        return data

    @classmethod
    def from_list(cls, data: list[dict]):
        components = cls()
        for row,action_row in enumerate(data):
            for component in action_row.get('components', []):
                components.add_component_raw(component, row)
            
        return components

def is_component(obj) -> bool:
    return isinstance(obj, (Button, Select, TextInput))


from typing import Union


class DiscordError(Exception):
    """Base class for all discord errors."""
    pass

class CogLoadError(DiscordError):
    """Raised when a cog cannot be loaded."""
    pass

class HTTPException(DiscordError):

    """Base class for all HTTP-related errors."""

    def __init__(self, message: Union[dict, str]) -> None:
        self.text: str = ''
        if isinstance(message, dict):
            self.text: str = f"{message.get('code', 0)}: {message.get('message', 'Unknown Error')}"
        else:
            self.text: str = message

class Forbidden(HTTPException):
    """Exception raised when a request is forbidden."""
    pass

class NotFound(HTTPException):
    """Exception raised when a resource is not found."""
    pass

class DiscordInternalError(HTTPException):
    """Exception raised when an internal error occurs."""
    pass


from dataclasses import dataclass, field
from datetime import datetime
from dacite import from_dict, config
from typing import Any, Dict, List, Optional, Union

from snowfin.components import Components

from .enums import ChannelType, OptionType, CommandType, ComponentType, Permissions, RequestType
from .embed import Embed

@dataclass
class Choice:
    """
    Class for the choices of an option.
    """
    name: str
    value: Union[str, int, float]

    def to_dict(self) -> Dict[str, Any]:
        return {
            "name": self.name,
            "value": self.value
        }

@dataclass
class User:
    id: int
    username: str
    discriminator: str
    avatar: str
    bot: Optional[bool]
    mfa_enabled: Optional[bool]
    banner: Optional[str]
    accent_color: Optional[str]
    locale: Optional[str]
    verified: Optional[bool]
    email: Optional[str]
    flags: Optional[int]
    premium_type: Optional[int]
    public_flags: Optional[int]

    @property
    def avatar_url(self) -> str:
        return f"https://cdn.discordapp.com/avatars/{self.id}/{self.avatar}.png"

@dataclass
class Member:
    user: Optional[User]
    nick: Optional[str]
    avatar: Optional[str]
    roles: list[int]
    joined_at: Optional[str]
    premium_since: Optional[str]
    deaf: Optional[bool]
    mute: Optional[bool]
    pending: Optional[bool]
    permissions: Optional[int]
    communication_disabled_until: Optional[str]

    def __post_init__(self):
        if self.joined_at:
            # get datetime from ISO8601 string
            self.joined_at = datetime.strptime(self.joined_at, '%Y-%m-%dT%H:%M:%S.%f+00:00')

        if self.premium_since:
            # get datetime from ISO8601 string
            self.premium_since = datetime.strptime(self.premium_since, '%Y-%m-%dT%H:%M:%S.%f+00:00')

        if self.communication_disabled_until:
            # get datetime from ISO8601 string
            self.communication_disabled_until = datetime.strptime(self.communication_disabled_until, '%Y-%m-%dT%H:%M:%S.%f+00:00')

    @property
    def avatar_url(self) -> str:
        if self.user is not None:
            return f"https://cdn.discordapp.com/avatars/{self.user.id}/{self.avatar}.png"

@dataclass
class RoleTags:
    bot_id: Optional[int]
    integration_id: Optional[int]
    premium_subscriber: Optional[bool]

@dataclass
class Role:
    id: int
    name: str
    color: int
    hoist: bool
    icon: Optional[str]
    unicode_emoji: Optional[str]
    position: int
    permissions: int
    managed: bool
    mentionable: bool
    tags: Optional[RoleTags]

@dataclass
class Channel:
    id: int
    name: str
    type: ChannelType
    permissions: int
    thread_metadata: Optional[Dict]
    parent_id: Optional[int]

@dataclass
class Component:
    """
    Discord command component
    """
    custom_id: Optional[str]
    component_type: Optional[ComponentType] # for when interaction data
    type: Optional[OptionType] # for when in a message
    values: Optional[List[str]] # for selects
    value: Optional[str] # for inputs
    label: Optional[str] # for inputs
    components: Optional[List['Component']] # for action rows
    style: Optional[int] # for non action rows

@dataclass
class Message:
    id: int
    channel_id: int
    guild_id: Optional[int]
    author: Optional[User | Member]
    member: Optional[Member]
    content: str
    timestamp: str
    edited_timestamp: Optional[str]
    tts: bool
    mention_everyone: bool
    mention_roles: list[int]
    mentions: list[User]
    attachments: list[dict]
    embeds: list[Embed]
    reactions: Optional[list[dict]]
    nonce: Optional[str | int]
    pinned: bool
    webhook_id: Optional[int]
    type: int
    activity: Optional[dict]
    application: Optional[dict]
    message_reference: Optional[dict]
    flags: Optional[int]
    referenced_message: Optional['Message']
    interaction: Optional[dict]
    thread: Optional[Channel]
    components: Optional[list[dict]]
    sticker_items: Optional[list[dict]]
    stickers: Optional[list[dict]]

    def __post_init__(self):
        if self.timestamp:
            # get datetime from ISO8601 string
            self.timestamp = datetime.strptime(self.timestamp, '%Y-%m-%dT%H:%M:%S.%f+00:00')

        if self.edited_timestamp:
            # get datetime from ISO8601 string
            self.edited_timestamp = datetime.strptime(self.edited_timestamp, '%Y-%m-%dT%H:%M:%S.%f+00:00')

        if self.components:
            self.components = Components.from_list(self.components)

@dataclass
class Resolved:
    users: Dict[str, User] = field(default_factory=dict)
    members: Dict[str, Member] = field(default_factory=dict)
    roles: Dict[str, Role] = field(default_factory=dict)
    channels: Dict[str, Channel] = field(default_factory=dict)
    messages: Dict[str, Message] = field(default_factory=dict)

    def get(self, type: OptionType, key: str | int) -> Any:
        key = str(key)
        if type is OptionType.USER:
            gotten = self.members.get(key, self.users.get(key))

            if isinstance(gotten, Member):
                gotten.user = self.users.get(key) or gotten.user

            return gotten

        elif type is OptionType.ROLE:
            return self.roles.get(key)
        elif type is OptionType.CHANNEL:
            return self.channels.get(key)
        elif type is OptionType.MENTIONABLE:
            gotten = self.members.get(key, self.users.get(key, self.roles.get(key)))

            if isinstance(gotten, Member):
                gotten.user = self.users.get(key) or gotten.user

            return gotten

@dataclass
class Option:
    focused: Optional[bool]
    name: str
    type: OptionType
    value: Optional[Union[str, int, float]]

@dataclass
class Command:
    id: int
    name: str
    guild_id: Optional[int]
    type: CommandType
    resolved: Optional[Resolved]
    options: List[Option] = field(default_factory=list)

@dataclass
class ModalSubmit:
    custom_id: str
    components: list[Component]

@dataclass
class Interaction:
    id: int
    application_id: int
    type: RequestType
    data: Optional[dict]
    guild_id: Optional[int]
    channel_id: Optional[int]
    member: Optional[Member]
    user: Optional[User]
    token: str
    version: int
    message: Optional[Message]
    local: Optional[str]
    guild_local: Optional[str]

    # added later
    client: Optional[Any]
    responded: bool = False

    def __post_init__(self) -> None:
        _config=config.Config(cast=[
            int,
            ChannelType,
            CommandType,
            OptionType,
            ComponentType,
            RequestType,
            Permissions
        ])

        if self.type in (RequestType.APPLICATION_COMMAND, RequestType.APPLICATION_COMMAND_AUTOCOMPLETE):
            self.data = from_dict(Command, self.data, config=_config)
        elif self.type is RequestType.MESSAGE_COMPONENT:
            self.data = from_dict(Component, self.data, config=_config)
        elif self.type is RequestType.MODAL_SUBMIT:
            self.data = from_dict(ModalSubmit, self.data, config=_config)
        else:
            raise ValueError(f'Unknown request type: {self.type}')

    @property
    def author(self) -> Member | User:
        return self.member if self.member else self.user


import asyncio
from email.mime import application
from typing import Any, Optional
import aiohttp
import json
import sanic
from dataclasses import asdict
from snowfin.decorators import SlashOption

from snowfin.response import _DiscordResponse
from snowfin.enums import CommandType

from .errors import *

__all__ = (
    'HTTP',
    'Route'
)

class Route:
    BASE: str = "https://discord.com/api/v10"

    def __init__(self, method: str, path: str, auth: bool = False, **params) -> None:
        self.method: str = method
        self.path: str = path
        self.url: str = self.BASE + self.path
        self.auth = auth
        self.params: dict = params

    def format(self, **extra_params) -> None:
        self.params.update(extra_params)
        for param,value in self.params.items():
            self.url = self.url.replace('{'+param+'}', str(value))

class BucketLock:
    """
    A lock for each ratelimit bucket
    """

    def __init__(self) -> None:
        self._lock: asyncio.Lock = asyncio.Lock()

        self.unlock_on_exit: bool = True
        self.limit: int = -1
        self.remaining: int = -1
        self.delta: float = 0.0

    def __repr__(self) -> str:
        return f"<BucketLock: {self.bucket_hash or 'Generic'}>"

    @property
    def locked(self) -> bool:
        """Return True if lock is acquired."""
        return self._lock.locked()

    def unlock(self) -> None:
        """Unlock this bucket."""
        self._lock.release()

    def ingest_ratelimit_header(self, header) -> None:
        """
        Ingests a discord rate limit header to configure this bucket lock.
        Args:
            header: A header from a http response
        """
        self.bucket_hash = header.get("x-ratelimit-bucket")
        self.limit = int(header.get("x-ratelimit-limit") or -1)
        self.remaining = int(header.get("x-ratelimit-remaining") or -1)
        self.delta = float(header.get("x-ratelimit-reset-after", 0.0))

    async def blind_defer_unlock(self) -> None:
        """Unlocks the BucketLock but doesn't wait for completion."""
        self.unlock_on_exit = False
        loop = asyncio.get_running_loop()
        loop.call_later(self.delta, self.unlock)

    async def defer_unlock(self) -> None:
        """Unlocks the BucketLock after a specified delay."""
        self.unlock_on_exit = False
        await asyncio.sleep(self.delta)
        self.unlock()

    async def __aenter__(self) -> None:
        await self._lock.acquire()

    async def __aexit__(self, *args) -> None:
        if self.unlock_on_exit and self._lock.locked():
            self.unlock()
        self.unlock_on_exit = True

class HTTP:
    """
    HTTP class
    """
    def __init__(
        self,
        application_id: int,
        token: Optional[str] = None,
        proxy: Optional[str] = None,
        proxy_auth: Optional[aiohttp.BasicAuth] = None,
        headers: Optional[dict] = None,
    ) -> None:
        self.application_id = application_id
        self.token = token
        self.proxy: Optional[str] = proxy
        self.proxy_auth: Optional[aiohttp.BasicAuth] = proxy_auth
        self.headers: dict = {
            "User-Agent": "Snowfin (https://github.com/kajdev/snowfin)",
        }

        self.locks = {}

        if headers is not None:
            self.headers.update(headers)

    def get_bucket(self, bucket_hash: str) -> BucketLock:
        """
        Get a bucket lock
        """
        if bucket_hash not in self.locks:
            self.locks[bucket_hash] = BucketLock()

        return self.locks[bucket_hash]

    async def request(
        self,
        route: Route,
        data: Optional[dict] = None,
        **kwargs
    ) -> Any:
        """
        Make a followup request
        """
        print(f"Requesting {route.method} {route.url} with {data}")

        headers = self.headers.copy()

        if 'headers' in kwargs:
            headers.update(kwargs.pop('headers'))

        if route.auth:
            if self.token is None:
                raise Exception(
                    "You must provide a token to make an authenticated request"
                )

            headers['Authorization'] = f"Bot {self.token}"

        route.format(application_id=self.application_id)

        async with aiohttp.ClientSession() as session:
            async with session.request(
                route.method,
                route.url,
                headers=headers,
                proxy=self.proxy,
                proxy_auth=self.proxy_auth,
                json=data if data is not None else None
            ) as response:

                response_data = None
                text = await response.text(encoding='utf-8')
                try:
                    if response.headers['content-type'] == 'application/json':
                        response_data = json.loads(text)
                except KeyError:
                    pass

                print(f"Got discord response ({response.status}): {response_data}")

                if 300 > response.status >= 200:
                    return response_data

                if response.status == 403:
                    raise Forbidden(response_data)
                elif response.status == 404:
                    raise NotFound(response_data)
                elif response.status >= 500:
                    raise DiscordInternalError(response_data)
                else:
                    raise HTTPException(response_data)

    async def close(self) -> None:
        """
        Close the HTTP session
        """
        await self._session.close()
    
    def send_followup(
        self,
        request: sanic.Request,
        response: _DiscordResponse,
        **kwargs
    ) -> Any:
        """
        Send a message
        """
        r = Route('POST', '/webhooks/{application_id}/{interaction_token}',
            interaction_token = request.ctx.token
        )

        return self.request(r,
            data=response.to_dict().get('data', {}),
            **kwargs
        )

    def edit_original_message(
        self,
        request: sanic.Request,
        response: _DiscordResponse,
        **kwargs
    ) -> Any:
        """
        Edit the original message
        """
        r = Route('PATCH', '/webhooks/{application_id}/{interaction_token}/messages/@original',
            interaction_token = request.ctx.token
        )

        data = response.to_dict().get('data', {})
        print(data)

        return self.request(r,
            data=data,
            **kwargs
        )

    def delete_original_message(
        self,
        request: sanic.Request,
        **kwargs
    ) -> Any:
        """
        Delete the original message
        """
        r = Route('DELETE', '/webhooks/{application_id}/{interaction_token}/messages/@original',
            interaction_token = request.ctx.token
        )

        return self.request(r, **kwargs)

    def edit_followup_message(
        self,
        request: sanic.Request,
        response: _DiscordResponse,
        message: int,
        **kwargs
    ) -> Any:
        """
        Edit the followup message
        """
        r = Route('PATCH', '/webhooks/{application_id}/{interaction_token}/messages/{message_id}',
            interaction_token = request.ctx.token,
            message_id = message
        )

        return self.request(r,
            data=response.to_dict().get('data', {}),
            **kwargs
        )

    def get_global_application_commands(
        self,
        **kwargs
    ) -> Any:
        """
        Get global application commands
        """
        r = Route('GET', '/applications/{application_id}/commands',
            auth=True
        )

        return self.request(r, **kwargs)

    def create_global_application_command(
        self,
        name: str,
        description: str,
        type: CommandType,
        name_localizations: Optional[dict] = None,
        description_localizations: Optional[dict] = None,
        options: Optional[list[Union[dict, SlashOption]]] = None,
        default_permission: bool = True,
        **kwargs
    ) -> Any:
        """
        Create global application command
        """
        r = Route('POST', '/applications/{application_id}/commands',
            auth=True
        )

        return self.request(r, 
            data={
                "name": name,
                "description": description,
                "type": type.value,
                "name_localizations": name_localizations,
                "description_localizations": description_localizations,
                "options": [asdict(o) if isinstance(o, SlashOption) else o for o in options],
                "default_permission": default_permission
            }
            **kwargs
        )

    def get_global_application_command(
        self,
        command_id: int,
        **kwargs
    ) -> Any:
        """
        Get global application command
        """
        r = Route('GET', '/applications/{application_id}/commands/{command_id}',
            command_id = command_id,
            auth=True
        )

        return self.request(r, **kwargs)

    def edit_global_application_command(
        self,
        command_id: int,
        name: Optional[str] = None,
        description: Optional[str] = None,
        name_localizations: Optional[dict] = None,
        description_localizations: Optional[dict] = None,
        options: Optional[list[Union[dict, SlashOption]]] = None,
        default_permission: Optional[bool] = None,
        **kwargs
    ) -> Any:
        """
        Edit global application command
        """
        r = Route('PATCH', '/applications/{application_id}/commands/{command_id}',
            command_id = command_id,
            auth=True
        )

        data = {}

        if name is not None:
            data['name'] = name

        if description is not None:
            data['description'] = description

        if name_localizations is not None:
            data['name_localizations'] = name_localizations

        if description_localizations is not None:
            data['description_localizations'] = description_localizations

        if options is not None:
            data['options'] = [asdict(o) if isinstance(o, SlashOption) else o for o in options]

        if default_permission is not None:
            data['default_permission'] = default_permission

        return self.request(r, 
            data=data
            **kwargs
        )

    def delete_global_application_command(
        self,
        command_id: int,
        **kwargs
    ) -> Any:
        """
        Delete global application command
        """
        r = Route('DELETE', '/applications/{application_id}/commands/{command_id}',
            command_id = command_id,
            auth=True
        )

        return self.request(r, **kwargs)

    def bulk_overwrite_global_application_commands(
        self,
        commands: list[dict],
        **kwargs
    ) -> Any:
        """
        Bulk overwrite global application commands
        """
        r = Route('PUT', '/applications/{application_id}/commands',
            auth=True
        )

        return self.request(r, 
            data=commands,
            **kwargs
        )

    def fetch_user(
        self,
        user_id: int,
        **kwargs
    ) -> Any:
        """
        Fetch current user
        """
        r = Route('GET', '/users/{user_id}',
            user_id = user_id,
            auth=True
        )

        return self.request(r, **kwargs)
    


from dataclasses import dataclass
from datetime import datetime
from typing import List, Union
from .color import Color

__all__ = (
    'Embed',
    'EmbedAuthor',
    'EmbedField',
    'EmbedFooter'
)

Empty = object()

@dataclass
class EmbedField:
    name: str
    value: str
    inline: bool = False

    def to_dict(self):
        return {
            'name': self.name,
            'value': self.value,
            'inline': self.inline,
        }

@dataclass
class EmbedAuthor:
    name: str
    url: str = Empty
    icon_url: str = Empty

    def to_dict(self):
        d = {
            'name': self.name,
        }
        if self.icon_url is not Empty:
            d['icon_url'] = self.icon_url
        if self.url is not Empty:
            d['url'] = self.url

        return d

@dataclass
class EmbedFooter:
    text: str
    icon_url: str = Empty

    def to_dict(self):
        d = {
            'text': self.text,
        }
        if self.icon_url is not Empty:
            d['icon_url'] = self.icon_url

        return d

@dataclass
class Embed:
    title: str = Empty
    description: str = Empty
    url: str = Empty
    color: Union[int, Color] = Empty
    timestamp: Union[int, datetime] = Empty
    footer: EmbedFooter = Empty
    image: str = Empty
    thumbnail: str = Empty
    author: EmbedAuthor = Empty
    fields: List[EmbedField] = Empty

    def add_field(self, name: str, value: str, inline: bool = False):
        if self.fields is Empty:
            self.fields = []

        self.fields.append(EmbedField(name, value, inline))

        return self

    def to_dict(self):
        d = {}

        for k, v in self.__dict__.items():
            if v is not Empty:
                if k == 'timestamp':
                    if isinstance(v, datetime):
                        v = int(v.timestamp())
                    else:
                        v = int(v)
                elif k == 'fields':
                    v = [x.to_dict() for x in v]
                elif k == 'color':
                    v = int(v)
                elif k in ('image', 'thumbnail'):
                    v = {'url': v}

                elif isinstance(v, (EmbedFooter, EmbedAuthor)):
                    v = v.to_dict()
            
                d[k] = v

        d['type'] = 'rich'
        return d